import os
import pathlib

from .metainfo import SPECIAL, POSSIBLE_QUALITY_TAGS, RATING_TAGS


tag_list_folder = pathlib.Path(os.path.dirname(__file__)) / "tag-list"
tag_lists = {
    os.path.splitext(f)[0]: set(open(tag_list_folder / f).read().strip().split("\n"))
    for f in os.listdir(tag_list_folder)
    if f.endswith(".txt")
}
tag_lists["special"] = set(SPECIAL)
tag_lists["quality"] = set(POSSIBLE_QUALITY_TAGS)
tag_lists["rating"] = set(RATING_TAGS)


def seperate_tags(all_tags):
    tag_map = {cate: [] for cate in tag_lists.keys()}
    tag_map["general"] = []
    for tag in all_tags:
        for cate in tag_lists.keys():
            if tag in tag_lists[cate]:
                tag_map[cate].append(tag)
                break
            if tag.replace("_", " ") in tag_lists[cate]:
                tag_map[cate].append(tag.replace("_", " "))
                break
        else:
            if len(tag) < 4:
                tag_map["general"].append(tag)
            else:
                tag_map["general"].append(tag.replace("_", " "))
    return tag_map


def apply_format(tag_map, form):
    for type in tag_map:
        if f"<|{type}|>" in form:
            if not tag_map[type]:
                form = form.replace(f"<|{type}|>,", "")
                form = form.replace(f"<|{type}|>", "")
            else:
                form = form.replace(f"<|{type}|>", ", ".join(tag_map[type]))
    return form.strip().strip(",")


def apply_dtg_prompt(tag_map, target="", aspect_ratio=1.0):
    special_tags = ", ".join(tag_map.get("special", []))
    rating = ", ".join(tag_map.get("rating", []))
    artist = ", ".join(tag_map.get("artist", []))
    characters = ", ".join(tag_map.get("characters", []))
    copyrights = ", ".join(tag_map.get("copyrights", []))
    general = ", ".join(tag_map.get("general", []))
    prompt = f"""
rating: {rating or '<|empty|>'}
artist: {artist.strip() or '<|empty|>'}
characters: {characters.strip() or '<|empty|>'}
copyrights: {copyrights.strip() or '<|empty|>'}
aspect ratio: {f"{aspect_ratio:.1f}" or '<|empty|>'}
target: {'<|' + target + '|>' if target else '<|long|>'}
general: {special_tags}, {general.strip().strip(",")}<|input_end|>
""".strip()

    return prompt


if __name__ == "__main__":
    from json import dumps

    print(tag_lists.keys())
    print([len(t) for t in tag_lists.values()])
    print(
        dumps(
            tag_map := seperate_tags(
                [
                    "1girl",
                    "fukuro daizi",
                    "kz oji",
                    "henreader",
                    "ask (askzy)",
                    "aki99",
                    "masterpiece",
                    "newest",
                    "absurdres",
                    "loli",
                    "solo",
                    "dragon girl",
                    "dragon horns",
                    "white dress",
                    "long hair",
                    "side up",
                    "river",
                    "tree",
                    "forest",
                    "pointy ears",
                    ":3",
                    "blue hair",
                    "blush",
                    "breasts",
                    "collarbone",
                    "dress",
                    "eyes visible through hair",
                    "fang",
                    "looking at viewer",
                    "nature",
                    "off shoulder",
                    "open mouth",
                    "orange eyes",
                    "tail",
                    "twintails",
                    "wings",
                ]
            ),
            ensure_ascii=False,
            indent=2,
        )
    )
    print()
    print(
        apply_format(
            tag_map,
            """<|special|>, 
<|characters|>, <|copyrights|>, 
<|artist|>, 

<|general|>, 

<|quality|>, <|meta|>, <|rating|>""",
        )
    )
    print()
    print()
    print(apply_dtg_prompt(tag_map, 1.0))



import regex as re
import random
from contextlib import nullcontext
from random import shuffle

import torch

try:
    from llama_cpp import Llama
except ImportError:

    class Llama:
        pass


from transformers import (
    GenerationConfig,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    set_seed,
)


def generate(
    model: PreTrainedModel | Llama,
    tokenizer: PreTrainedTokenizerBase,
    prompt="",
    temperature=0.5,
    top_p=0.95,
    top_k=45,
    repetition_penalty=1.17,
    max_new_tokens=128,
    autocast_gen=lambda: torch.autocast("cpu", enabled=False),
    **kwargs,
):
    if isinstance(model, Llama):
        # print(kwargs)
        result = model.create_completion(
            prompt,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            max_tokens=max_new_tokens,
            repeat_penalty=repetition_penalty or 1,
            seed=kwargs.get("seed", None),
        )
        # print(prompt, "===", result["choices"][0]["text"])
        return prompt + result["choices"][0]["text"]
    if "seed" in kwargs:
        set_seed(kwargs["seed"])

    torch.cuda.empty_cache()
    inputs = tokenizer(prompt, return_tensors="pt")
    input_ids = inputs["input_ids"].to(next(model.parameters()).device)
    generation_config = GenerationConfig(
        temperature=temperature,
        top_p=top_p,
        top_k=top_k,
        repetition_penalty=repetition_penalty,
        do_sample=True,
        **kwargs,
    )
    with torch.no_grad(), autocast_gen():
        generation_output = model.generate(
            input_ids=input_ids,
            generation_config=generation_config,
            return_dict_in_generate=True,
            output_scores=True,
            max_new_tokens=max_new_tokens,
        )
    s = generation_output.sequences[0]
    output = tokenizer.decode(s)

    torch.cuda.empty_cache()
    return output


def black_list_match(tag, black_list):
    for b in black_list:
        if re.match(b, tag):
            return True
    return False


def tag_gen(
    text_model,
    tokenizer,
    prompt,
    prompt_tags,
    len_target,
    black_list,
    temperature=0.5,
    top_p=0.95,
    top_k=100,
    max_new_tokens=256,
    max_retry=20,
    max_same_output=15,
    seed=None,
):
    retry = max_retry
    llm_gen = ""

    set_seed(seed)

    iter_count = 0
    prev_output = set()
    same_output_count = 0
    while retry >= 0 and same_output_count < max_same_output:
        llm_gen = generate(
            model=text_model,
            tokenizer=tokenizer,
            prompt=prompt,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            repetition_penalty=None,
            max_new_tokens=max_new_tokens,
            stream_output=False,
            autocast_gen=nullcontext,
            prompt_lookup_num_tokens=10,
            pad_token_id=getattr(tokenizer, "eos_token_id", None),
            eos_token_id=getattr(tokenizer, "eos_token_id", None),
            seed=seed + iter_count,
        )
        iter_count += 1
        llm_gen = llm_gen.replace("</s>", "").replace("<s>", "")
        orig_prompt = llm_gen.split("<|input_end|>")[0]
        extra = llm_gen.split("<|input_end|>")[-1].strip().strip(",")
        extra_tokens = sorted(
            set(
                [
                    tok.strip()
                    for tok in extra.split(",")
                    if not black_list_match(tok.strip(), black_list)
                ]
            )
        )
        llm_gen = llm_gen.replace(extra, ", ".join(extra_tokens))

        yield llm_gen, extra_tokens, iter_count

        if set(extra_tokens) == prev_output:
            same_output_count += 1
            retry += 1
        else:
            same_output_count = 0
            prev_output = set(extra_tokens)

        if len(prompt_tags) + len(extra_tokens) < len_target:
            retry -= 1
            shuffle(extra_tokens)
            llm_gen = f"{orig_prompt}<|input_end|>{', '.join(extra_tokens)}"
            prompt = llm_gen.strip().replace("  <|", " <|")
        else:
            break
    yield llm_gen, extra_tokens, iter_count



SPECIAL = [
    "1girl",
    "2girls",
    "3girls",
    "4girls",
    "5girls",
    "6+girls",
    "multiple girls",
    "1boy",
    "2boys",
    "3boys",
    "4boys",
    "5boys",
    "6+boys",
    "multiple boys",
    "male focus",
    "1other",
    "2others",
    "3others",
    "4others",
    "5others",
    "6+others",
    "multiple others",
]
POSSIBLE_QUALITY_TAGS = [
    "masterpiece",
    "best quality",
    "great quality",
    "good quality",
    "normal quality",
    "low quality",
    "worse quality",
    "very aesthetic",
    "aesthetic",
    "displeasing",
    "very displeasing",
    "newest",
    "recent",
    "mid",
    "early",
    "old",
    "score_9",
    "score_8_up",
    "score_7_up",
    "score_6_up",
    "score_5_up",
    "score_4_up",
    "source_anime",
    "source_cartoon",
    "source_furry",
    "source_pony",
]
RATING_TAGS = ["safe", "general", "sfw", "sensitive", "nsfw", "explicit"]
TARGET = {
    "very_short": 8,
    "short": 16,
    "long": 30,
    "very_long": 45,
}



from setuptools import setup, find_packages


setup(
    name="tipo-kgen",
    packages=find_packages(),
    version="0.0.4",
    license="Apache 2.0",
    url="https://github.com/KohakuBlueleaf/KGen",
    description=(
        "TIPO: Text to Image genration through "
        "text Presampling with LLMs for Optimal prompting"
    ),
    author="Shih-Ying Yeh(KohakuBlueLeaf)",
    author_email="apolloyeh0123@gmail.com",
    zip_safe=False,
    install_requires=[
        "transformers",
        "huggingface_hub",
        "torch",
    ],
    package_data={"": ["**/*.txt"]},
)



import os
import pathlib

from huggingface_hub import hf_hub_download
from transformers import LlamaForCausalLM, LlamaTokenizer

from .logging import logger


model_dir = pathlib.Path(__file__).parent / "models"
model_list = [
    "KBlueLeaf/DanTagGen-alpha",
    "KBlueLeaf/DanTagGen-beta",
    "KBlueLeaf/DanTagGen-gamma",
]
gguf_name = [
    "ggml-model-f16.gguf",
    "ggml-model-Q8_0.gguf",
    "ggml-model-Q6_K.gguf",
]


try:
    from llama_cpp import Llama
except Exception:
    logger.warning("Llama-cpp-python cannot be imported")
    Llama = None


if not os.path.isdir(model_dir):
    os.makedirs(model_dir, exist_ok=True)


def download_gguf(model_name="KBlueLeaf/DanTagGen-gamma", gguf_name=gguf_name[-1]):
    logger.info(f"Downloading gguf model from {model_name}")
    result = hf_hub_download(
        model_name,
        gguf_name,
        repo_type="model",
        cache_dir=None,
        local_dir=model_dir,
        local_dir_use_symlinks=False,
    )
    new_name = model_dir / f"{model_name.split('/')[-1]}_{gguf_name}"
    if os.path.isfile(new_name):
        os.remove(new_name)
    os.rename(result, new_name)
    logger.info(f"Downloaded gguf model to {new_name}")
    return new_name


def list_gguf():
    files = [
        str(model_dir / file)
        for file in os.listdir(model_dir)
        if file.endswith(".gguf")
    ]
    return files


def load_model(model_name="KBlueLeaf/DanTagGen-gamma", gguf=False):
    global text_model, tokenizer
    if gguf:
        try:
            assert Llama is not None
            model_name = os.path.basename(model_name)
            text_model = Llama(
                str(model_dir / model_name),
                n_ctx=384,
                n_gpu_layers=100,
                verbose=False,
            )
            tokenizer = None
            logger.info(f"Llama-cpp-python/gguf model {model_name} loaded")
            return
        except Exception as e:
            logger.warning(f"Llama-cpp-python/gguf model {model_name} load failed")
            model_name = model_list[-1]
    logger.info(f"Using transformers model {model_name}")
    text_model = LlamaForCausalLM.from_pretrained(model_name).eval().half()
    tokenizer = LlamaTokenizer.from_pretrained(model_name)
    logger.info(f"Model {model_name} loaded")


if __name__ == "__main__":
    model_file = download_gguf()
    load_model(model_file, gguf=True)
